# Duo, A crazy eights game.
# Copyright (C) 2009 The Duo developers.
#
# This file is part of Duo.
#
# Duo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""The protocol level of game messaging service.

The Duo protocol is the way for players (or AIs) and the referee to
communicate inside a game. It defines messages for all the game
events, and only them; the connexion to a game or the media used by
the communication are not covered by the Duo protocol and are the task
of the transport layer (`duo.transport`).

Along with its implicit meaning, all messages have an implicit
direction. Some of these messages can only be emitted by the referee
and sent to one player, or broadcasted to all of them. Others are only
sent by the players to another one, or to the referee. To represent
these directions throughout this module documentation, *P* will stand
for player, *R* for referee and *A* for all players (meaning a
broadcasted message).

A typical game setup looks like the following.

::

      Player A        Player B        Player C            Referee
     (connection)                                           |
          |                                                 |
          |                                     <= AUTH(A) =|
          |          (connection)                           |
          |               |                                 |
          |               |                   <= AUTH(A,B) =|
          |               |          (connection)           |
          |               |               |                 |
          |               |               | <= AUTH(A,B,C) =|
          |               |               |                 |
          |------------------ START (key) ----------------->|
          |               |               |                 |


The following is part of the chosen ruleset. In the case of Duo
default rules, the referee deals the cards, and waits everyone is
ready. When this is the case, it returns the first card, compute the
new actions for everyone and send new Hand messages.

::

      Player A        Player B        Player C            Referee
          |               |               |                 |
          |<---------- HAND (cards, [ReadyAction]) ---------|
          |-------------- ACTION (ReadyAction) ------------>|
          |               |               |                 |
                                (...)
          |               |               |                 |
          |               |    <= HEAP (card, color, None) =|
          |               |                    <= TURN (A) =|
          |<-------------- HAND(cards, [...]) --------------|
          |               |               |                 |
"""

from __future__ import with_statement

__docformat__ = 'restructuredtext'

import logging
import threading

from duo import transport


class Message(object):
    """The Duo protocol elementary piece of information.

    A Duo message is implicitly defined by its type, and the
    information it carries is completed by its payload. The payload
    vary with the kind of information carried by the message. If more
    than one object is attached to the message, the payload is a
    tuple.
    """
    def __init__(self, payload=None):
        self._payload = payload

    def __str__(self):
        return self.__class__.__name__ + ': ' + str(self._payload)

    def __repr__(self):
        return self.__class__.__name__ + ': ' + str(self._payload)

    def _get_payload(self):
        """Retrieve the content of this message."""
        return self._payload

    payload = property(_get_payload, None, None, 'The message payload')


# From referee
class AuthMessage(Message):
    """Message broadcasted by the referee each time a player joins the game.

    There is no authentication at the Duo protocol level.
    Authentication must be achieved by the underlying transport
    layer. It must ensure for example that two players always have a
    different name, or that there is no mismatch on connections.

    Every time a player joins a game, the game referee broadcasts a
    `duo.protocol.AuthMessage` to all the player of the game. This
    message provides in its payload the updated list of the game
    players.

    ::

      R->A: AUTH <list of players>

    This way, all players are informed of the new comer. This one also
    gets the name of all the previous players connected and the one of
    the Referee.
    """
    def __init__(self, players_list):
        Message.__init__(self, players_list)


class DealMessage(Message):
    """Message broadcasted by the referee when a player receives new cards.

    This message payload is a tuple with the number of cards drawn and
    the recipient name. The message is meant to warn all the players
    of this event. The player who drew the cards receives the Hand
    message which provides his full hand.

    ::

      R->A: DEAL <nb of cards, player>
    """
    def __init__(self, nb_cards, player):
        Message.__init__(self, (nb_cards, player))


class HandMessage(Message):
    """Message sent by the referee to players to update their hand status.

    This message is emitted each time an event occurs in the game so
    that players can know what are their possible moves. The message
    payload is a tuple with the new full hand of the player, and a set
    containing all the actions he can undertake at this instant.

    The actions set is specially important for AI which, this way, do
    not have to embed the game rules to play. It also can help the
    player's interface to provide hints.

    ::

      R->P: HAND <cards, actions>
    """
    def __init__(self, hand, actions):
        Message.__init__(self, (hand, actions))


class HeapMessage(Message):
    """Message broadcasted by the referee each time the heap changes.

    The message payload is a tuple with the new heap, the demanded
    color, and the player who played the card (None being the referee itself).

    ::

      R->A: HEAP <card, color, player>
    """
    def __init__(self, card, color, player):
        Message.__init__(self, (card, color, player))


class TurnMessage(Message):
    """Message broadcasted by the referee each time a new player should play.

    The message payload is the name of the player who should now
    play. Note that, depending on the rules, this can only be a hint
    on the game status. In the end, the actions sent to a player in
    the HandMessage are the only source of valid moves.

    ::

      R->A: TURN <player>
    """
    def __init__(self, player):
        Message.__init__(self, player)


class ReverseMessage(Message):
    """Message broadcasted by the referee when a Reverse card is played.

    There is no payload attached to this message.

    ::

      R->A: REVERSE
    """
    def __init__(self):
        Message.__init__(self)


class WinnerMessage(Message):
    """Message broadcasted by the referee when the game came to an end.

    The message payload is the name of the winner.

    ::

      R->A: WINNER <player>
    """
    def __init__(self, player):
        Message.__init__(self, player)


# From players
class StartMessage(Message):
    """Message emitted by a player to start the game.

    The player must provide a key, which will be checked by the
    referee. The game will only start if the provided key matches the
    key used to create the referee.

    ::

      P->R: START <key>
    """
    def __init__(self, key):
        Message.__init__(self, key)


class ActionMessage(Message):
    """Message emitted by a player to proceed.

    The message payload is the action undertaken.

    ::

      P->R: ACTION <action>
    """
    def __init__(self, action):
        Message.__init__(self, action)


# Any way
class ErrorMessage(Message):
    """Message emitted by an entity on error.

    The message payload is the exception raised, or the error
    string. Currently, this message is only used by the referee.
    """
    def __init__(self, exception):
        Message.__init__(self, exception)


class MessageBundle(list):
    """A group of message."""
    pass


class PlayerAlreadyAuthenticatedError(Exception):
    """Error raised on multiple player authentication."""
    def __init__(self):
        Exception.__init__(self)


class ProtocolAdapter(object):
    """The interface to the transport layer dedicated to the Duo protocol.

    The ProtocolAdapter provides basic registering to the transport
    layer, and Duo messages mangling and demangling. It is neutral in
    the sense that it does not provide any high level service which
    could be associated to a player or a referee. This is the role of
    inherited classes.
    """
    def __init__(self, name):
        self._name = name
        self._hooks = {}
        self._endpoint = None

    def connect(self, transport_layer, transport_options):
        """Connect to the transport layer."""
        transport_options['callbacks'] = {
            transport.TransportJoinPacket: self.on_endpoint_join,
            transport.TransportEscapePacket: self.on_endpoint_escape,
            transport.ProtocolPacket: self.on_message }

        self._endpoint = transport.create_endpoint(
            transport_layer, transport_options, self.name)

    def is_connected(self):
        """Return true if connected, false otherwise."""
        return self._endpoint != None

    def register(self, msg_class, callback, argument=None):
        """Register a new callback."""
        t = msg_class.__name__
        logging.getLogger('duo.protocol').debug(
            '%s registered %s with %s' % (self.name, t, callback.__name__))
        self._hooks[t] = (callback, argument)

    def on_message(self, message, emitter):
        """Callback invoked by the transport layer on new message."""
        if isinstance(message, MessageBundle):
            for msg in message:
                self._notify(msg, emitter)
        else:
            self._notify(message, emitter)

    def send_to(self, message, recipient):
        """Send a message to one recipient."""
        logging.getLogger('duo.protocol').debug(
            '%s sends %s to %s' % (self.name, message, recipient))
        self._endpoint.send_to(message, recipient)

    def broadcast(self, message):
        """Send a message to all the recipients in the same context."""
        logging.getLogger('duo.protocol').debug(
            '%s broadcasts: %s' % (self.name, message))
        self._endpoint.broadcast(message)

    def on_endpoint_join(self, message, emitter):
        """Callback invoked by the transport layer on new connection."""
        pass

    def on_endpoint_escape(self, message, emitter):
        """Callback invoked by the transport layer on disconnection."""
        pass

    def _get_name(self):
        """Retrieve the name of this adapter."""
        return self._name

    name = property(_get_name, None, None, "The Protocol Adapter name")

    def _notify(self, message, emitter):
        """Dispatch the reveiced message to the registered callbacks."""
        msg_type = message.__class__.__name__
        try:
            callback, arg = self._hooks[msg_type]
            logging.getLogger('duo.protocol').debug(
                'notifying %s to %s with %s'
                % (msg_type, self.name, callback.__name__))
            callback.__call__(message.payload, emitter, arg)
        except KeyError, ex:
            logging.getLogger('duo.protocol').debug(
                'ProtocolAdapter._notify no callback for %s in %s' %
                (message.__class__.__name__, ex))
        except Exception, ex:
            logging.getLogger('duo.protocol').error(
                'ProtocolAdapter._notify %s: %s' %
                (message.__class__.__name__, ex))


class Player(ProtocolAdapter):
    """The player class from a user point of view.

    Any user interface needs to work with at least one Player instance;
    the one representing the user for the game. There should also be one
    instance for each one of the AI players running on the machine.

    The Player interface provides the minimal but strictly necessary
    methods to manage a player representation.
    """
    def __init__(self, name, cb_argument=None, ai=None):
        ProtocolAdapter.__init__(self, name)

        self._hand_lock = threading.Lock()
        self._hand = []
        self._actions = set()
        self._referee = None
        self._ai = ai
        if self._ai:
            self._ai.initialize(self)

        # Connect essential callbacks for a player.
        self.register(AuthMessage, self.on_auth_cb, cb_argument)
        self.register(HandMessage, self.on_hand_cb, cb_argument)
        self.register(TurnMessage, self.on_turn_cb, cb_argument)
        self.register(HeapMessage, self.on_heap_cb, cb_argument)
        self.register(DealMessage, self.on_deal_cb, cb_argument)
        self.register(ErrorMessage, self.on_error_cb, cb_argument)
        self.register(WinnerMessage, self.on_winner_cb, cb_argument)

    def is_robot(self):
        """Return True if this player is an automaton."""
        return self._ai != None

    def send_to_referee(self, message):
        """Send the given message to the referee only."""
        self.send_to(message, self._referee)

    def get_games(self):
        """Retrieve the available games in the neighbourhood."""
        return self._endpoint.get_contexts()

    def join_game(self, name):
        """Join the game whose name is given."""
        self._endpoint.join_context(name)

    def leave_game(self):
        """Leave the current game."""
        self._endpoint.escape()
        if self._ai:
            self._ai.stop_game()

    def _get_hand(self):
        """Retrieve the current cards of the player."""
        with self._hand_lock:
            return self._hand

    def _set_hand(self, cards):
        """Update the cards of the player."""
        with self._hand_lock:
            self._hand = cards

    hand = property(_get_hand, _set_hand, None, 'The player hand')

    def _get_actions(self):
        """Retrieve the possible actions of the player."""
        with self._hand_lock:
            return self._actions

    actions = property(_get_actions, None, None, 'The player possible actions')

    def start_game(self, key=None):
        """Ask the referee to deal the cards."""
        if self._ai:
            self._ai.start_game()
        else:
            self.send_to_referee(StartMessage(key))

    def do(self, action):
        """Undertake an action."""
        self.send_to_referee(ActionMessage(action))

    def on_auth_cb(self, payload, emitter, args):
        """Hook called on authentication."""
        self._referee = emitter
        if self._ai:
            self._ai.on_auth(payload, emitter, args)

    def on_hand_cb(self, payload, emitter, args):
        """Hook called on new hand."""
        with self._hand_lock:
            self._hand, self._actions = payload
        if self._ai:
            self._ai.on_hand(payload, emitter, args)

    def on_turn_cb(self, payload, emitter, args):
        """Hook called on new turn."""
        if self._ai:
            self._ai.on_turn(payload, emitter, args)

    def on_heap_cb(self, payload, emitter, args):
        """Hook called on new heap."""
        if self._ai:
            self._ai.on_heap(payload, emitter, args)

    def on_deal_cb(self, payload, emitter, args):
        """Hook called when a new deal is done, whoever the player."""
        if self._ai:
            self._ai.on_deal(payload, emitter, args)

    def on_error_cb(self, payload, emitter, args):
        """Hook called on error."""
        if self._ai:
            self._ai.on_error(payload, emitter, args)

    def on_winner_cb(self, payload, emitter, args):
        """Hook called on game end."""
        if self._ai:
            self._ai.on_winner(payload, emitter, args)
