# Duo, A crazy eights game.
# Copyright (C) 2004,2009 The Duo developers.
#
# This file is part of Duo.
#
# Duo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""A Tk interface for the Duo game.

The Tk interface is the original game interface. The interface code
was isolated in the duo.tk module to become only a choice among
others like ncurses or GTK based interfaces.

Each interface brings its own limitations. This interface was designed
for two to four players. Player hands are not limited, but the game
can become somewhat difficult when the number of cards become very
important.
"""

# Paths for an improved game experience:
#
#   - Cards placing should be optimized: raise can be used more
#     specifically, and useless card movement for AIs shall be
#     removed.
#
#   - Ready and Pass buttons should be bigger. Keyboard shortcuts
#     should be cool too (like space to pass...).
#
#   - Taking care of the resize event should be nice. But is it
#     possible to scale on the fly the whole canvas with its cards ?
#

from __future__ import with_statement

import os
import logging
import threading

import Tkinter as tk
import tkMessageBox as tk_msg
import tkSimpleDialog as tk_dlg
import Pmw
import PIL.Image
import PIL.ImageTk
#import PIL.PngImagePlugin

import duo
from duo import ai
from duo import game
from duo import protocol
from duo import referee
from duo import tools
from duo import ui


# Constants.
DEFAULT_WINDOW_WIDTH = 500
DEFAULT_WINDOW_HEIGHT = 400

CARDSET = 'classic'

MOVE_STEPS = 20
MOVE_SPEED = 20
ORDER_SPEED = 5


ABOUT_DIALOG_INFO = u"""
Duo is an open source card game based on the original Uno game.

Current Maintainer:
David Soulayrol <david.soulayrol@gmail.com>

Original Author:
Camille Roux <windoz@tiscali.fr>

Website:
http://duo.tuxfamily.org/
http://launchpad.net/duo/"""



# How to sort the hand.
# Could vary on user choice...
COLORS_ORDER = ['Y', 'G', 'R', 'B']
def compare(card, other_card):
    result = 0

    # Compare no colored cards.
    if card.color == 'N':
        if not other_card.color != 'N':
            result = game.Card.TYPES.index(card.type).__cmp__(
                game.Card.TYPES.index(other_card.type))
        else:
            result = -1
    elif other_card.color == 'N':
        result = 1
    else:
        result = COLORS_ORDER.index(card.color).__cmp__(
            COLORS_ORDER.index(other_card.color))
        if result == 0:
            result = game.Card.TYPES.index(card.type).__cmp__(
                game.Card.TYPES.index(other_card.type))
    return result


class Graphics(object):
    def __init__(self, default_x_size, default_y_size):
        self._width = default_x_size
        self._height= default_y_size

    width = property(lambda self: self._width)
    height = property(lambda self: self._height)
    card_width = property(lambda self: (self._height / 5) * 2/3)
    card_height = property(lambda self: self._height / 5)
    card_space = property(lambda self: (self._height / 5 ) / 3)
    card_x_border = property(lambda self: ((self._height / 5) * 2/3 ) / 3)
    card_y_border = property(lambda self: ((self._height / 5) * 2/3 ) / 3)


class Card:
    """UI Card object.

    A UI card (in Tk UI) knows how to load its image, and how to move
    or rotate itself. Its movement can be invoked with different
    animation speeds or can be immediate.
    """

    TAG_VISIBLE = 'visible'
    TAG_BACK = 'back'

    def __init__(self, root, card, player, x=0, y=0):
        """Constructor.

        A brand new UI card always embeds a protocol card and a
        position. It immediately loads its image depending on these
        two information bits. Optionnaly, it can be given an initial
        position, its default beging the center of the carpet.
        """
        gr = root.get_graphics()

        self._root = root
        self._x = x if x else gr.width / 2 + gr.card_space + gr.card_width / 2
        self._y = y if y else gr.height / 2
        self._height = gr.card_height if not player % 2 else gr.card_width
        self._width = gr.card_width if not player % 2 else gr.card_height
        self._player = player
        self._card = card
        self._image = None
        self._image_cache = None
        self._image_tags = None

        self._draw()

    def __repr__(self):
        return repr(self._card)

    def __str__(self):
        return str(self._card)

    def impl(self):
        """The underlying protocol representation of the card."""
        return self._card

    def is_turned_down(self):
        """Whether the card face is visible."""
        return self._player != Player.SOUTH and self._player != Player.DECK

    def get_card_file(self):
        """Build the file representing this card using its name."""
        name_root = ''
        if self._card.type == 'D':
            name_root = '+2-' \
                        + game.Card.COLORS[self._card.color].lower()
        elif self._card.type == 'R':
            name_root = 'inv-' \
                        + game.Card.COLORS[self._card.color].lower()
        elif self._card.type == 'P':
            name_root = 'pas-' \
                        + game.Card.COLORS[self._card.color].lower()
        elif self._card.type == 'W':
            name_root = 'jok'
        elif self._card.type == 'X':
            name_root = '+4'
        else:
            name_root = self._card.type + '-' \
                        + game.Card.COLORS[self._card.color].lower()

        return os.path.join('images', CARDSET, name_root + ".png")

    def place(self, player, order, move_speed=0):
        """Move this UI card instance.

        This method is used to move a card into a player's hand
        only. It accepts the player's position, and the ordered
        position in the player's hand. The move is immediate, unless a
        speed value is given.
        """
        gr = self._root.get_graphics()
        x, y = 0, 0
        if player == Player.SOUTH:
            x = order * gr.card_space + gr.card_x_border
            y = gr.height - self._height - gr.card_y_border

        elif player == Player.WEST:
            x = gr.card_y_border
            y = gr.card_x_border + order * gr.card_space

        elif player == Player.NORTH:
            x = gr.width - gr.card_width - order * gr.card_space - gr.card_x_border
            y = gr.card_y_border

        elif player == Player.EAST:
            x = gr.width - gr.card_height - gr.card_y_border
            y = gr.height - gr.card_width - gr.card_x_border - order * gr.card_space

        self._move(x, y, move_speed)

    def play(self, after_cb=None):
        """Play this UI card instance.

        This method is used to move a card towards the heap only. The
        movement is always animated. If the card to be played was
        hidden - because it was the card of one of the player's
        opponents, it is first returned.
        """
        if self.is_turned_down():
            self.suppr()
            self._draw(turn_up=True)

        gr = self._root.get_graphics()
        self._move(gr.width / 2 - gr.card_width,
                   gr.height / 2 - gr.card_height / 2,
                   move_speed=MOVE_SPEED, after_cb=after_cb)

    def suppr(self):
        """Remove this card from the canvas."""
        self._root.get_canvas().delete(self._image)

    def _draw(self, turn_up=False):
        """Draw or redraw the card on the canvas.

        If turn_up is set, then the card is drawn with the figure
        visible, whatever its position. It is used in particular to
        redraw an opponent card before it is played.
        """
        # TODO: ON RUN...
        gr = self._root.get_graphics()
        canvas = self._root.get_canvas()
        self._image_tags = Card.TAG_VISIBLE

        file = ''
        if self._card is None or not turn_up and self.is_turned_down():
            card_file = os.path.join('images', CARDSET, 'back.png')
            self._image_tags = Card.TAG_BACK
        else:
            card_file = self.get_card_file()
            self._image_tags = Card.TAG_VISIBLE

        try:
            img = PIL.Image.open(card_file)
            if self._player % 2:
                if turn_up:
                    self._width, self._height = self._height, self._width
                else:
                    img = img.rotate(90 * self._player)
            img = img.resize((self._width, self._height), PIL.Image.ANTIALIAS)

            self._image_cache = PIL.ImageTk.PhotoImage(img)
            self._image = canvas.create_image(
                self._x , self._y, image = self._image_cache, tags=self._image_tags)
        except Exception, e:
            print e
            tk_msg.showwarning(
                _('Error'), _("You didn't install the card set correctly"))

    def _move(self, x, y, move_speed=0, step=0, after_cb=None):
        """The recursive animation movement method."""
        canvas = self._root.get_canvas()
        if step == 0:
            canvas.tkraise(self._image)
            canvas.tag_raise(Card.TAG_VISIBLE, Card.TAG_BACK)

        if move_speed == 0 or step == MOVE_STEPS:
            self._x, self._y = x, y
            canvas.coords(self._image, x + self._width / 2, y + self._height / 2)
            if after_cb:
                after_cb()

        elif move_speed != 0:
            i = MOVE_STEPS - step - 0.9
            self._x = x + (i/MOVE_STEPS) * (self._x - x)
            self._y = y + (i/MOVE_STEPS) * (self._y - y)
            canvas.coords(self._image,
                          self._x + self._width / 2,
                          self._y + self._height / 2)
            canvas.after(
                move_speed, self._move, x, y, move_speed, step + 1, after_cb)


class StartDialog(tk_dlg.Dialog):
    """The start game dialog.

    This dialog asks the player for its preferences related to the
    game to come (the number of players, the number of cards or the
    ruleset to use).
    """

    PLAYERS_NB = [2, 3, 4]
    CARDS_NB = range(1, 16)

    def __init__(self, parent, model):
        # _model must be defined first since it is used in body, and
        # body is called by the constructor.
        self._model = model
        self._validated_model = None
        tk_dlg.Dialog.__init__(self, parent, _('New game'))

    def body(self, master):
        self._name_entry = Pmw.EntryField(
            master,
            labelpos=tk.W,
            label_text=_('Your name:'),
            validate={'max': 10})
        self._name_entry.pack(anchor=tk.W)
        self._name_entry.insert(0, self._model[ui.OPT_GAME_PLAYER_NAME])

        self._nb_players_menu = Pmw.OptionMenu(
            master,
            labelpos=tk.W,
            label_text=_('Number of players:'),
            items=StartDialog.PLAYERS_NB,
            menubutton_width=3)
        self._nb_players_menu.pack(anchor=tk.W, pady=5)
        self._nb_players_menu.setvalue(self._model[ui.OPT_GAME_NB_PLAYERS])

        self._nb_cards_menu = Pmw.ComboBox(
            master,
            labelpos=tk.W,
            label_text=_('Number of cards for each player:'),
            scrolledlist_items=StartDialog.CARDS_NB,
            entryfield_validate={'validator': 'integer', 'min': '1', 'max': '15'})
        self._nb_cards_menu.pack(anchor=tk.W)
        self._nb_cards_menu.selectitem(
            StartDialog.CARDS_NB.index(self._model[ui.OPT_GAME_NB_CARDS]))

        self._ruleset_menu = Pmw.OptionMenu(
            master,
            labelpos=tk.W,
            label_text=_('Ruleset:'),
            items=list(tools.get_modules('duo.rules')),
            menubutton_width=10)
        self._ruleset_menu.pack(anchor=tk.W, pady=5)
        self._ruleset_menu.setvalue(self._model[ui.OPT_GAME_RULESET])

        Pmw.alignlabels(
            [self._nb_players_menu, self._nb_cards_menu,
             self._ruleset_menu, self._name_entry])

    def buttonbox(self):
        """Add button box.

        Default button box is overriden so as to handle localisation.
        """
        box = tk.Frame(self)

        self._ok_button = tk.Button(box, text=_('Ok'), width=10, command=self.ok)
        self._ok_button.pack(side=tk.LEFT, padx=5, pady=5)

        tk.Button(box, text=_('Cancel'), width=10, command=self.cancel).pack(
            side=tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack()

    def validate(self):
        """Validate the data.

        This method is called automatically to validate the data
        before the dialog is destroyed. In our case, most of the
        validation is already handled by the mega widgets. Definitive
        checking is done and model is updated.
        """
        try:
            self._model[ui.OPT_GAME_PLAYER_NAME] = self._name_entry.getvalue()
            self._model[ui.OPT_GAME_NB_PLAYERS] = int(
                self._nb_players_menu.getcurselection())
            self._model[ui.OPT_GAME_NB_CARDS] = int(self._nb_cards_menu.get())
            self._model[ui.OPT_GAME_RULESET] = self._ruleset_menu.getcurselection()
            self._validated_model = self._model
            return True
        except ValueError:
            return False

    def _get_model(self):
        return self._validated_model

    model = property(_get_model, None, None, 'Get the resulting data.')


class LanguageDialog(tk_dlg.Dialog):
    """The language selection dialog.

    Duo selects automatically the user locale if the translation is
    available on start. This dialog allows the user to change it the
    time the program runs; the selection is not saved between game
    session.
    """

    def __init__(self, parent, values, default):
        # _values and _lang must be defined first since they are used
        # in body, and body is called by the constructor.
        self._values = values
        self._default = default
        self._lang = None
        tk_dlg.Dialog.__init__(self, parent, _(u'Language choice'))

    def body(self, master):
        self._menu = Pmw.OptionMenu(
            master,
            labelpos=tk.W,
            label_text=_('Please select your language:'),
            items=self._values,
            menubutton_width=6,
            command=self._on_selected_cb)

        self._menu.setvalue(self._default)
        self._menu.pack()

        return self._menu.component('menubutton')

    def buttonbox(self):
        """Add button box.

        Default button box is overriden so as to handle localisation
        and to grey the Ok button by default.
        """
        box = tk.Frame(self)

        self._ok_button = tk.Button(box, text=_('Ok'), width=10, command=self.ok)
        self._ok_button.configure(state=tk.DISABLED)
        self._ok_button.pack(side=tk.LEFT, padx=5, pady=5)

        tk.Button(box, text=_('Cancel'), width=10, command=self.cancel).pack(
            side=tk.LEFT, padx=5, pady=5)

        self.bind("<Escape>", self.cancel)

        box.pack()

    def validate(self):
        """Validate the data.

        This method is called automatically to validate the data
        before the dialog is destroyed. In our case, it simply stores
        the selected language.
        """
        self._lang = self._menu.getcurselection()
        return True

    def _on_selected_cb(self, value):
        """React to the option menu selection.

        On selection, the Ok button is activated and focused, and the
        Return key is bound to the ok event.
        """
        if self._ok_button.cget('state') == tk.DISABLED:
            self._ok_button.configure(state=tk.ACTIVE)
            self._ok_button.focus_set()
            self.bind('<Return>', self.ok)

    def _get_lang(self):
        return self._lang

    lang = property(_get_lang, None, None, 'Get the selected language.')


class Player(protocol.Player):
    SOUTH, WEST, NORTH, EAST, DECK = range(5)

    def __init__(self, name, transport, transport_options, ui, position=SOUTH, ai=None):
        protocol.Player.__init__(self, name, ai=ai)
        logging.getLogger('duo.ui.tk').debug('Player.__init__ %s', (name))

        self._ui = ui
        self._position = position
        self._cards = []
        self._cards_lock = threading.Lock()

        # TODO: Use OptionSets
        # Options should be managed from command line or
        # through the interface. Fixed for now.
        self.connect(transport, transport_options)

        if not self.is_robot():
            self._ui.get_canvas().bind('<Button-1>', self._on_mouse_event)
            self._ui.get_canvas().bind_all(
                '<KeyPress-space>', self._on_space_key_event)

    def prepare(self):
        self.hand = []
        with self._cards_lock:
            self._cards = []

    def get_position(self):
        return self._position

    def find_card(self, impl):
        card_found = None
        with self._cards_lock:
            for card in self._cards:
                if card.impl() == impl:
                    card_found = card
                    break
        return card_found

    def draw(self):
        self.do(referee.DrawAction(self.name))

    def try_play(self, card):
        logging.getLogger('duo.ui.tk').debug(
            'Player.tryPlay: local player(%s) is playing %s' % (self.name, card))
        color = card.impl().color
        if color == 'N':
            color = self._ui.get_color_choice()
        self.do(referee.CardAction(self.name, card.impl(), color))

    def on_hand_cb(self, message, emitter, args):
        previous_hand = self.hand
        protocol.Player.on_hand_cb(self, message, emitter, args)
        self._ui.defer(self._on_hand, self.hand, previous_hand)

    def on_turn_cb(self, message, emitter, args):
        protocol.Player.on_turn_cb(self, message, emitter, args)

        if message == self.name:
            self._ui.defer(self._ui.update_turn_line, self)

        # Whoever the turn, reset the player's draw button.
        self._ui.defer(self._ui.get_pass_button().config, state=tk.DISABLED)

    def on_heap_cb(self, message, emitter, args):
        protocol.Player.on_heap_cb(self, message, emitter, args)
        (card, color, player) = message
        self._ui.defer(self._on_heap, card, color, player)

    def on_error_cb(self, message, emitter, args):
        protocol.Player.on_error_cb(self, message, emitter, args)
        # TODO: Display dialog error ?
        print '## Error: ' + str(message)

    def on_winner_cb(self, message, emitter, args):
        protocol.Player.on_winner_cb(self, message, emitter, args)

        # Do just once:
        if message == self.name:
            self._ui.after(1000, self._ui.end_game, message)

    def _on_mouse_event(self, event):
        """Mouse events handling.

        This method is called inside the main Tk loop.
        """
        gr = self._ui.get_graphics()
        with self._cards_lock:
            nb_cards = len(self._cards)

        # Remember origin is the upper left corner.
        y_bottom = gr.height - gr.card_y_border
        y_top = y_bottom  - gr.card_height
        x_left = gr.card_x_border
        x_right = x_left + (nb_cards - 1) * gr.card_space \
                  + gr.card_width

        deck_bottom = gr.height / 2 + gr.card_height / 2
        deck_top = gr.height / 2 - gr.card_height / 2
        deck_left = gr.width / 2 + gr.card_space
        deck_right = deck_left + gr.card_width

        if event.y < y_bottom and event.y > y_top \
               and event.x > x_left and event.x < x_right:
            idx = min((event.x - x_left) / gr.card_space, nb_cards - 1)
            with self._cards_lock:
                self.try_play(self._cards[idx])

        elif event.y < deck_bottom and event.y > deck_top \
                 and event.x > deck_left and event.x < deck_right:
            self.draw()

    def _on_space_key_event(self, event):
        """The magical space key behaviour.

        The space key is attached to both ready and pass buttons. It
        can be used to activate the first of these buttons which is
        active.
        """
        buttons = [self._ui.get_ready_button(), self._ui.get_pass_button()]
        buttons = [b for b in buttons if b.cget('state') == tk.NORMAL]
        if buttons:
            buttons[0].invoke()

    def _on_hand(self, hand, previous_hand):
        """Actual hand notification handling.

        This method is called inside the main Tk loop.
        """
        logging.getLogger('duo.ui.tk').debug(
            'Player._on_hand: new hand %s for %s' % (hand, self.name))

        # AI do not need sorting cards ;)
        if not self.is_robot():
            hand.sort(compare)

        # TODO: For the moment, ignore the action updates. This is
        # achieved only after sort, or player hand will never be the
        # same.
        if hand == previous_hand:
            return

        idx = 0
        if previous_hand:
            # Compare the new hand and the previous one to find out
            # if a card from this hand was played.
            #
            # NOTE: this is incorrect in case of two identical cards
            # in the hand, because the rightest one is always chosen.
            if len(hand) == len(previous_hand) - 1:
                idx = len(hand) - 1
                while idx > 0:
                    if hand[idx] != previous_hand[idx + 1]:
                        break;
                    idx -= 1
            # ... or if cards were drawn.
            elif not self.is_robot():
                self._ui.defer(
                    self._ui.get_pass_button().config, state=tk.NORMAL)

        self._ui.defer(self.refresh_cards, hand, idx)

        # AI is always ready. Human player must react himself.
        if not previous_hand:
            if not self.is_robot():
                self._ui.defer(
                    self._ui.get_ready_button().config, state=tk.NORMAL)

    def _on_heap(self, card, color, player):
        """Actual heap notification handling.

        This method is called inside the main Tk loop.
        """
        logging.getLogger('duo.ui.tk').debug(
            'Player._on_heap: new heap is %s(%s) by %s' % (card, color, player))
        if player == self.name:
            tk_card = self.find_card(card)
            self._ui.play_card(tk_card)
            self._ui.update_label(color)
            with self._cards_lock:
                self._cards.remove(tk_card)

        elif not self.is_robot() and player is None:
            self._ui.set_heap_card(card)

    def refresh_cards(self, hand, from_index):
        logging.getLogger('duo.ui.tk').debug('Player.refresh_cards %s', self.name)
        # TODO: Manage differently to speed up.
        # - While placing card, raise should not be used when card is reused.
        #    ... hmph... and on draw ?
        with self._cards_lock:
            current_impls = [c.impl() for c in self._cards]
            tk_cards = []
            tk_card = None
            for idx, card in enumerate(hand):
                # Create of reuse the Tk card.
                if card in current_impls:
                    tk_card = self._cards[current_impls.index(card)]
                    current_impls.remove(card)
                    self._cards.remove(tk_card)
                    if idx >= from_index:
                        tk_card.place(self._position, idx, ORDER_SPEED)
                else:
                    tk_card = Card(self._ui, card, self._position)
                    tk_card.place(self._position, idx, MOVE_SPEED)
                tk_cards.append(tk_card)
            self._cards = tk_cards

    def __str__(self):
        return self._player.name


class UserInterface(tk.Tk):
    def __init__(self, options):
        tk.Tk.__init__(self)
        logging.getLogger('duo.ui').debug('tk.UserInterface.__init__')

        # Asynchronous Events Management.
        self._lock = threading.Lock()
        self._events = []

        self._game = None
        self._player = None
        self._graphics = None
        self._heap_card = None
        self._deck_card = None
        self._turn_line = None
        self._color_lbl = None

        self._defaults = ui.sanitize_options(options)

        self._y_message = tk.StringVar(self)
        self._g_message = tk.StringVar(self)
        self._r_message = tk.StringVar(self)
        self._b_message = tk.StringVar(self)
        self._pass_message = tk.StringVar(self)
        self._ready_message = tk.StringVar(self)

        self._color_var = tk.StringVar(self)
        self._color_var.set('Y')

        self._create()
        self.wm_resizable(0, 0)

    def launch(self):
        """The interface entry point.

        The duo starter script gives control to the interface using
        this method, as soon as it has been built. This implementation
        checks if an argument was passed on the command line to start
        the game immediately.
        """
        if self._defaults[ui.OPT_GAME_START]:
            self.after(500, self.start_default_game)
        self.mainloop()

    def defer(self, callable_, *args, **keys):
        """Defer an event to the interface main loop.

        All graphics related events (canvas drawing, button actions
        and so on) MUST be handled by the Tk mainloop. This method can
        be used by any object to queue an event to this loop, whatever
        the thread it is running in."""
        with self._lock:
            self._events.insert(0, (callable_, args, keys))

    def _create(self):
        """Draw the application window."""
        #Main window
        Pmw.initialise(self)
        self.title(_(u'DUO'))
        self._canvas = tk.Canvas(self, bg = 'dark green',
                                 width = DEFAULT_WINDOW_WIDTH,
                                 height = DEFAULT_WINDOW_HEIGHT)
        self._canvas.pack(side = tk.LEFT, fill = tk.BOTH, expand = 1)

        frame = tk.Frame(self)

        self._info_lbl = tk.Label(frame)
        self._info_lbl.pack(side = tk.TOP, fill = tk.Y, expand = 1)

        frame.pack(expand=1, side=tk.TOP)

        frame = tk.Frame(self, borderwidt=1, relief=tk.GROOVE)

        tk.Radiobutton(
            frame, anchor=tk.W, textvariable=self._y_message,
            selectcolor='#f8f500', value='Y', variable=self._color_var
            ).pack(side=tk.TOP, fill=tk.X)

        tk.Radiobutton(
            frame, anchor=tk.W, textvariable=self._g_message,
            selectcolor='#21d909', value='G', variable=self._color_var
            ).pack(side=tk.TOP, fill=tk.X)

        tk.Radiobutton(
            frame, anchor=tk.W, textvariable=self._r_message,
            selectcolor='#f50000', value='R', variable=self._color_var
            ).pack(side=tk.TOP, fill=tk.X)

        tk.Radiobutton(
            frame, anchor=tk.W, textvariable=self._b_message,
            selectcolor='#236ad5', value='B', variable=self._color_var
            ).pack(side=tk.TOP, fill=tk.X)

        frame.pack(side=tk.TOP, fill=tk.X)

        self._pass_button = tk.Button(
            self, textvariable=self._pass_message, height=3, font=18,
            command=self._pass_cb, state=tk.DISABLED)
        self._pass_button.pack(side = tk.BOTTOM, fill = tk.X)

        self._ready_button = tk.Button(
            self, textvariable=self._ready_message, height=3, font=18,
            command=self._ready_cb, state=tk.DISABLED)
        self._ready_button.pack(side = tk.BOTTOM, fill = tk.X)

        self._update_localisation()

        # Menu
        self._create_menu()

        # End of the program
        self.protocol("WM_DELETE_WINDOW", self._on_menu_quit_cb)

        # Start events polling on canvas.
        self._canvas.after(100, self._poll_canvas_events)

    def _create_menu(self):
        """Draw or redraw the application menu."""
        menu = tk.Menu(self)
        self.config(menu = menu)
        game_menu = tk.Menu(menu)
        menu.add_cascade(label = _('Game'), menu = game_menu)
        game_menu.add_command(label = _('New game'), command = self._on_menu_start_cb)
        game_menu.add_separator()
        game_menu.add_command(label = _('Quit'), command = self._on_menu_quit_cb)

        option_menu = tk.Menu(menu)
        menu.add_cascade(label = _('Options'), menu = option_menu)
        option_menu.add_command(label=_('Language'), command = self._on_menu_language_cb)

        help_menu = tk.Menu(menu)
        menu.add_cascade(label = _('Help'), menu = help_menu)
        help_menu.add_command(label = _('About...'), command = self._on_menu_about_cb)

    def _update_localisation(self):
        """Redraw labels taking care of localisation.

        Note that menus can't be built using text variables as far as
        I know, so this method doesn't handle menus.
        """
        self._y_message.set(_('Yellow'))
        self._g_message.set(_('Green'))
        self._r_message.set(_('Red'))
        self._b_message.set(_('Blue'))
        self._pass_message.set( _('Pass') )
        self._ready_message.set( _('Ready') )

    def create_game(self, options):
        logging.getLogger('duo.ui').debug('tk.UserInterface._create_game')

        name = options[ui.OPT_GAME_PLAYER_NAME]
        transport = options[ui.OPT_GAME_TRANSPORT]

        # Prepare the board
        self._graphics = Graphics(self._canvas.winfo_width(),
                                  self._canvas.winfo_height())

        # Create Game
        # TODO: create of get coordinates in function of protocol
        #lServerCoordinates = ('', 8881)
        self._game = ui.Game(name + '_party')
        transport_options = { 'context': self._game.name }

        referee_options = {
            'cards': options[ui.OPT_GAME_NB_CARDS],
            'ruleset': options[ui.OPT_GAME_RULESET],
            'master_key': self._game.key
            }
        self._game.referee = referee.Referee(
            transport, transport_options, referee_options)

        # Create player.
        self._player = Player(name, transport, transport_options, self)
        self._game.add_player(self._player)

        # Then eventually create the local players instances, with their
        # correct position
        names = []
        for i in range(options[ui.OPT_GAME_NB_PLAYERS] - 1):
            names.append(name)
            robot, name = ai.load('camille', names)
            self._game.add_player(
                Player(name, transport, transport_options, self, i + 1, robot))

        # Initialize points
        self.update_scores()

    def start_game(self):
        logging.getLogger('duo.ui').debug('tk.UserInterface.start_game')

        # Draw the current player
        self._turn_line = self._canvas.create_line(
            0, 0, 0, 0, fill = 'yellow', width = 5, smooth = False)

        self._color_lbl = self._canvas.create_text(
            self._graphics.width / 2 - self._graphics.card_width / 2,
            self._graphics.height / 2 + self._graphics.card_height / 2 + 10,
            text = '', fill = 'yellow', font = 15)

        # Put the Deck
        x = self._graphics.width / 2 + self._graphics.card_space \
             + self._graphics.card_width / 2
        y = self._graphics.height / 2
        self._deck_card = Card(self, None, Player.DECK, x = x, y = y)

        # Let the human player be the last one to start, so that
        # all AI can initialize their stuff.
        for player in self._game.get_players():
            player.prepare()
            if player != self._player:
                player.start_game()
        self._player.start_game(self._game.key)

    def start_default_game(self):
        """Start a game without question to the user, using default values.
        """
        self.create_game(self._defaults)
        self.start_game()

    def end_game(self, winner):
        if winner == self._player.name:
            message = _('Congratulations %s, you won!\n\n') % winner
        else:
            message =_('%s won.\n\n') % winner

        scores = self.compute_round_scores()
        message += _('Round:\n')
        for player in self._game.get_players():
            message += _('%s: %d points\n') % (
                player.name, scores[player])

        self.update_scores(scores)
        message += _('\nGame:\n')
        for player in self._game.get_players():
            message += _('%s: %d points\n') % (
                player.name, self._game.get_score(player.name))

        message += _('\nDo you want to continue the game?')
        if tk_msg.askyesno(_("WIN"),message):
            self._canvas.delete(tk.ALL)
            self.start_game()
        else:
            self._canvas.delete(tk.ALL)

    def leave(self):
        if self._game:
            self._game.end()
        self.destroy()

    def get_canvas(self):
        return self._canvas

    def get_graphics(self):
        return self._graphics

    def get_player(self, name):
        return self._game.get_player(name)

    def get_heap_card(self):
        return self._heap_card

    def get_color_choice(self):
        return self._color_var.get()

    def play_card(self, card):
        logging.getLogger('duo.ui.tk').debug('UserInterface.play_card')

        # The clean_heap function is not bound until the play_card
        # method exists. So the top_card variable is necessary to keep
        # a reference to the current heap card because the _heap_card
        # member will be overwritten before the method exits.
        top_card = self._heap_card
        def clean_heap():
            if top_card:
                top_card.suppr()

        card.play(after_cb=clean_heap)
        self._heap_card = card

    def set_heap_card(self, impl):
        """Place a card from the deck on the heap.

        This movement is essentially used to put the first card on the
        heap at game start."""
        logging.getLogger('duo.ui.tk').debug('UserInterface.set_heap_card')
        card = Card(self, impl, Player.DECK)
        self.play_card(card)

    def get_pass_button(self):
        return self._pass_button

    def get_ready_button(self):
        return self._ready_button

    def update_label(self, color):
        """Draw the color label on the canvas."""
        logging.getLogger('duo.ui.tk').debug('UserInterface.update_label')
        if self._heap_card.impl().type in ['W','X']:
            self._canvas.itemconfig(
                self._color_lbl, text = _(game.Card.COLORS[color]))
        else:
            self._canvas.itemconfig(self._color_lbl, text = '')

    def update_turn_line(self, player):
        """Draw the turn indicator.

        The turn indicator is the yellow line drawn on the side of one player.
        """
        logging.getLogger('duo.ui.tk').debug('UserInterface.update_turn_line')
        if player.get_position() == Player.SOUTH:
            self._canvas.coords(
                self._turn_line, 5, self._graphics.height - 5,
                self._graphics.width - 5, self._graphics.height - 5)
        elif player.get_position() == Player.WEST:
            self._canvas.coords(
                self._turn_line, 5, 5, 5, self._graphics.height - 5)
        elif player.get_position() == Player.NORTH:
            self._canvas.coords(
                self._turn_line, self._graphics.width - 5, 5, 5, 5)
        elif player.get_position() == Player.EAST:
            self._canvas.coords(
                self._turn_line, self._graphics.width - 5,
                self._graphics.height - 5, self._graphics.width - 5, 5)

    def update_scores(self, scores=None):
        """Draw the scores on the window side."""
        logging.getLogger('duo.ui.tk').debug('UserInterface.updateScores')
        label = ''
        for player in self._game.get_players():
            if scores is None:
                label += '%s: %d\n' % (player.name, 0)
            else:
                self._game.add_to_score(player.name, scores[player])
                label += '%s: %d\n' % (
                    player.name, self._game.get_score(player.name))
        self._info_lbl['text'] = label

    def compute_round_scores(self):
        scores = {}
        for player in self._game.get_players():
            scores[player] = 0
            for card in player.hand:
                scores[player] += game.Card.POINTS[card.type]
        return scores

    def _on_menu_start_cb(self):
        d = StartDialog(self, self._defaults)
        if d.model:
            self._canvas.delete(tk.ALL)
            self.create_game(d.model)
            self.start_game()

    def _on_menu_quit_cb(self):
        """Callback invoked to quit the game.

        This method asks for confirmation before the program is really
        stopped."""
        if tk_msg.askokcancel(_("Quit"), _("Do you want to quit this program?")):
            self.leave()

    def _on_menu_language_cb(self):
        """Callback invoked to change the interface language.

        This method first display the language dialog to let the user
        choose its new language. If no translation can be found, a
        warning dialog is built instead.

        When the user validates his choice, gettext is invoked and
        interface elements are refreshed."""
        try:
            d = LanguageDialog(
                self, sorted(os.listdir("locale")), duo.get_default_language())
            if d.lang:
                duo.switch_to_language(d.lang)

                # Rebuild interface elements to update language.
                #  ... Canvas is not rebuilt.
                self._create_menu()
                self._update_localisation()

        except OSError:
            self.title()
            tk_msg.showwarning(
                _(u'No languages found.'),
                u'Please use make-locale.py to generate the translations.')

    def _on_menu_about_cb(self):
        """Callback invoked to display the About dialog."""
        tk_msg.showinfo(_('About'), _(ABOUT_DIALOG_INFO))

    def _poll_canvas_events(self):
        """The canvas event polling routine.

        This callback is invoked every 100ms when the game is launched
        (from _create). It polls any pending events deferred to the
        interface with the defer method."""
        # TODO: detect end.
        with self._lock:
            count = len(self._events)
        while count > 0:
            with self._lock:
                callable_, args, keys = self._events.pop()
            if callable(callable_):
                callable_(*args, **keys)
            count -= 1
        self._canvas.after(100, self._poll_canvas_events)

    def _pass_cb(self):
        """Callback invoked on pass action from the player."""
        self._pass_button.config(state = tk.DISABLED)
        self._player.do(referee.PassAction(self._player.name))

    def _ready_cb(self):
        """Callback invoked on ready action from the player."""
        self._ready_button.config(state = tk.DISABLED)
        self._player.do(referee.ReadyAction(self._player.name))
